# -*- coding: utf-8 -*-

from __future__ import absolute_import

import six
import json
import datetime
import collections

import enum
import decorator
from six.moves import http_client as httplib

from betfair import exceptions
from betfair.meta import utils
from betfair.meta.models import BetfairModel


def get_chunks(sequence, chunk_size):
    """Split sequence into chunks.

    :param list sequence:
    :param int chunk_size:
    """
    return [
        sequence[idx:idx + chunk_size]
        for idx in range(0, len(sequence), chunk_size)
    ]


def get_kwargs(kwargs):
    """Get all keys and values from dictionary where key is not `self`.

    :param dict kwargs: Input parameters
    """
    return {
        key: value for key, value in six.iteritems(kwargs)
        if key != 'self'
    }


def check_status_code(response, codes=None):
    """Check HTTP status code and raise exception if incorrect.

    :param Response response: HTTP response
    :param codes: List of accepted codes or callable
    :raises: ApiError if code invalid
    """
    codes = codes or [httplib.OK]
    checker = (
        codes
        if callable(codes)
        else lambda resp: resp.status_code in codes
    )
    if not checker(response):
        raise exceptions.ApiError(response, response.json())


def result_or_error(response):
    """Get `result` field from Betfair response or raise exception if not
    found.

    :param Response response:
    :raises: ApiError if no results passed
    """
    data = response.json()
    result = data.get('result')
    if result is not None:
        return result
    raise exceptions.ApiError(response, data)


def process_result(result, model=None):
    """Cast response JSON to Betfair model(s).

    :param result: Betfair response JSON
    :param BetfairModel model: Deserialization format; if `None`, return raw
        JSON
    """
    if model is None:
        return result
    if isinstance(result, collections.Sequence):
        return [model(**item) for item in result]
    return model(**result)


class BetfairEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        if isinstance(o, BetfairModel):
            o.validate()
            return o.serialize()
        if isinstance(o, enum.Enum):
            return o.name
        return super(BetfairEncoder, self).default(o)


def make_payload(base, method, params):
    """Build Betfair JSON-RPC payload.

    :param str base: Betfair base ("Sports" or "Account")
    :param str method: Betfair endpoint
    :param dict params: Request parameters
    """
    payload = {
        'jsonrpc': '2.0',
        'method': '{base}APING/v1.0/{endpoint}'.format(**locals()),
        'params': utils.serialize_dict(params),
        'id': 1,
    }
    return payload


@decorator.decorator
def requires_login(func, *args, **kwargs):
    """Decorator to check that the user is logged in. Raises `BetfairError`
    if instance variable `session_token` is absent.
    """
    self = args[0]
    if self.session_token:
        return func(*args, **kwargs)
    raise exceptions.NotLoggedIn()



# -*- coding: utf-8 -*-

from __future__ import division

from decimal import Decimal, ROUND_HALF_UP


CUTOFFS = (
    (2, 100),
    (3, 50),
    (4, 20),
    (6, 10),
    (10, 5),
    (20, 2),
    (30, 1),
    (50, 0.5),
    (100, 0.2),
    (1000, 0.1),
)


MIN_PRICE = 1.01
MAX_PRICE = 1000


def as_dec(value):
    return Decimal(str(value))


def arange(start, stop, step):
    while start < stop:
        yield start
        start += step


def make_prices(min_price, cutoffs):
    prices = []
    cursor = as_dec(min_price)
    for cutoff, step in cutoffs:
        prices.extend(arange(as_dec(cursor), as_dec(cutoff), as_dec(1 / step)))
        cursor = cutoff
    prices.append(as_dec(MAX_PRICE))
    return prices


PRICES = make_prices(MIN_PRICE, CUTOFFS)


def nearest_price(price, cutoffs=CUTOFFS):
    """Returns the nearest Betfair odds value to price.

    Adapted from Anton Zemlyanov's AlgoTrader project (MIT licensed).
    https://github.com/AlgoTrader/betfair-sports-api/blob/master/lib/betfair_price.js

    :param float price: Approximate Betfair price (i.e. decimal odds value)
    :param tuple cutoffs: Optional tuple of (cutoff, step) pairs
    :returns: The nearest Befair price
    :rtype: float
    """
    if price <= MIN_PRICE:
        return MIN_PRICE
    if price > MAX_PRICE:
        return MAX_PRICE

    price = as_dec(price)
    for cutoff, step in cutoffs:
        if price < cutoff:
            break
    step = as_dec(step)
    return float((price * step).quantize(2, ROUND_HALF_UP) / step)


def ticks_difference(price_1, price_2):
    """Returns the absolute difference in terms of "ticks" (i.e. individual
    price increments) between two Betfair prices.

    :param float price_1: An exact, valid Betfair price
    :param float price_2: An exact, valid Betfair price
    :returns: The absolute value of the difference between the prices in "ticks"
    :rtype: int
    """
    price_1_index = PRICES.index(as_dec(price_1))
    price_2_index = PRICES.index(as_dec(price_2))
    return abs(price_1_index - price_2_index)


def price_ticks_away(price, n_ticks):
    """Returns an exact, valid Betfair price that is n_ticks "ticks" away from
    the given price. n_ticks may positive, negative or zero (in which case the
    same price is returned) but if there is no price n_ticks away from the
    given price then an exception will be thrown.

    :param float price: An exact, valid Betfair price
    :param float n_ticks: The number of ticks away from price the new price is
    :returns: An exact, valid Betfair price
    :rtype: float
    """
    price_index = PRICES.index(as_dec(price))
    return float(PRICES[price_index + n_ticks])



# -*- coding: utf-8 -*-


class BetfairError(Exception):
    pass


class NotLoggedIn(BetfairError):
    pass


class LoginError(BetfairError):

    def __init__(self, response, data):
        self.response = response
        self.status_code = response.status_code
        self.message = data.get('loginStatus', 'UNKNOWN')
        super(LoginError, self).__init__(self.message)


class AuthError(BetfairError):

    def __init__(self, response, data):
        self.response = response
        self.status_code = response.status_code
        self.message = data.get('error', 'UNKNOWN')
        super(AuthError, self).__init__(self.message)


class ApiError(BetfairError):

    def __init__(self, response, data):
        self.response = response
        self.status_code = response.status_code
        try:
            error_data = data['error']['data']['APINGException']
            self.message = error_data.get('errorCode', 'UNKNOWN')
            self.details = error_data.get('errorDetails')
        except KeyError:
            self.message = 'UNKNOWN'
            self.details = None
        super(ApiError, self).__init__(self.message)



# -*- coding: utf-8 -*-

import six
import inflection
from schematics import types
from schematics import models


class BetfairModelMeta(models.ModelMeta):
    """Set default `serialized_name` and `deserialize_from` of Schematics types
    to camel-cased attribute names.
    """
    def __new__(meta, name, bases, attrs):
        for name, attr in six.iteritems(attrs):
            if isinstance(attr, types.BaseType):
                camelized = inflection.camelize(name, uppercase_first_letter=False)
                attr.serialized_name = attr.serialized_name or camelized
                attr.deserialize_from = attr.deserialize_from or camelized
        return super(BetfairModelMeta, meta).__new__(meta, name, bases, attrs)


class BetfairModel(six.with_metaclass(BetfairModelMeta, models.Model)):

    def __init__(self, **data):
        super(BetfairModel, self).__init__()
        self.import_data(data)

    def import_data(self, data, **kwargs):
        kwargs['strict'] = False
        return super(BetfairModel, self).import_data(data, **kwargs)



# -*- coding: utf-8 -*-

"""BetFair betting enums. See
https://api.developer.betfair.com/services/webapps/docs/display/1smk3cen4v3lu3yomq5qye0ni/Betting+Enums
"""

from enum import Enum


MarketProjection = Enum(
    'MarketProjection', [
        'COMPETITION',
        'MARKET_DESCRIPTION',
        'EVENT',
        'EVENT_TYPE',
        'RUNNER_METADATA',
        'RUNNER_DESCRIPTION',
        'MARKET_START_TIME',
    ]
)

PriceData = Enum(
    'PriceData', [
        'SP_AVAILABLE',
        'SP_TRADED',
        'EX_BEST_OFFERS',
        'EX_ALL_OFFERS',
        'EX_TRADED',
    ]
)

MatchProjection = Enum(
    'MatchProjection', [
        'NO_ROLLUP',
        'ROLLED_UP_BY_PRICE',
        'ROLLED_UP_BY_AVG_PRICE',
    ]
)

OrderProjection = Enum(
    'OrderProjection', [
        'ALL',
        'EXECUTABLE',
        'EXECUTION_COMPLETE',
    ]
)

MarketStatus = Enum(
    'MarketStatus', [
        'INACTIVE',
        'OPEN',
        'SUSPENDED',
        'CLOSED',
    ]
)

RunnerStatus = Enum(
    'RunnerStatus', [
        'ACTIVE',
        'WINNER',
        'LOSER',
        'REMOVED_VACANT',
        'REMOVED',
        'HIDDEN',
    ]
)

TimeGranularity = Enum(
    'TimeGranularity', [
        'DAYS',
        'HOURS',
        'MINUTES',
    ]
)

Side = Enum(
    'Side', [
        'BACK',
        'LAY',
    ]
)

OrderStatus = Enum(
    'OrderStatus', [
        'EXECUTION_COMPLETE',
        'EXECUTABLE',
    ]
)

OrderBy = Enum(
    'OrderBy', [
        'BY_BET',
        'BY_MARKET',
        'BY_MATCH_TIME',
        'BY_PLACE_TIME',
        'BY_SETTLED_TIME',
        'BY_VOID_TIME',
    ]
)

SortDir = Enum(
    'SortDir', [
        'EARLIEST_TO_LATEST',
        'LATEST_TO_EARLIEST',
    ]
)

OrderType = Enum(
    'OrderType', [
        'LIMIT',
        'LIMIT_ON_CLOSE',
        'MARKET_ON_CLOSE',
    ]
)

MarketSort = Enum(
    'MarketSort', [
        'MINIMUM_TRADED',
        'MAXIMUM_TRADED',
        'MINIMUM_AVAILABLE',
        'MAXIMUM_AVAILABLE',
        'FIRST_TO_START',
        'LAST_TO_START',
    ]
)

MarketBettingType = Enum(
    'MarketBettingType', [
        'ODDS',
        'LINE',
        'RANGE',
        'ASIAN_HANDICAP_DOUBLE_LINE',
        'ASIAN_HANDICAP_SINGLE_LINE',
        'FIXED_ODDS',
    ]
)

ExecutionReportStatus = Enum(
    'ExecutionReportStatus', [
        'SUCCESS',
        'FAILURE',
        'PROCESSED_WITH_ERRORS',
        'TIMEOUT',
    ]
)

ExecutionReportErrorCode = Enum(
    'ExecutionReportErrorCode', [
        'ERROR_IN_MATCHER',
        'PROCESSED_WITH_ERRORS',
        'BET_ACTION_ERROR',
        'INVALID_ACCOUNT_STATE',
        'INVALID_WALLET_STATUS',
        'INSUFFICIENT_FUNDS',
        'LOSS_LIMIT_EXCEEDED',
        'MARKET_SUSPENDED',
        'MARKET_NOT_OPEN_FOR_BETTING',
        'DUPLICATE_TRANSACTION',
        'INVALID_ORDER',
        'INVALID_MARKET_ID',
        'PERMISSION_DENIED',
        'DUPLICATE_BETIDS',
        'NO_ACTION_REQUIRED',
        'SERVICE_UNAVAILABLE',
        'REJECTED_BY_REGULATOR',
    ]
)

PersistenceType = Enum(
    'PersistenceType', [
        'LAPSE',
        'PERSIST',
        'MARKET_ON_CLOSE',
    ]
)

InstructionReportStatus = Enum(
    'InstructionReportStatus', [
        'SUCCESS',
        'FAILURE',
        'TIMEOUT',
    ]
)

InstructionReportErrorCode = Enum(
    'InstructionReportErrorCode', [
        'INVALID_BET_SIZE',
        'INVALID_RUNNER',
        'BET_TAKEN_OR_LAPSED',
        'BET_IN_PROGRESS',
        'RUNNER_REMOVED',
        'MARKET_NOT_OPEN_FOR_BETTING',
        'LOSS_LIMIT_EXCEEDED',
        'MARKET_NOT_OPEN_FOR_BSP_BETTING',
        'INVALID_PRICE_EDIT',
        'INVALID_ODDS',
        'INSUFFICIENT_FUNDS',
        'INVALID_PERSISTENCE_TYPE',
        'ERROR_IN_MATCHER',
        'INVALID_BACK_LAY_COMBINATION',
        'ERROR_IN_ORDER',
        'INVALID_BID_TYPE',
        'INVALID_BET_ID',
        'CANCELLED_NOT_PLACED',
        'RELATED_ACTION_FAILED',
        'NO_ACTION_REQUIRED',
    ]
)

RollupModel = Enum(
    'RollupModel', [
        'STAKE',
        'PAYOUT',
        'MANAGED_LIABILITY',
        'NONE',
    ]
)

GroupBy = Enum(
    'GroupBy', [
        'EVENT_TYPE',
        'EVENT',
        'MARKET',
        'SIDE',
        'BET',
    ]
)

BetStatus = Enum(
    'BetStatus', [
        'SETTLED',
        'VOIDED',
        'LAPSED',
        'CANCELLED',
    ]
)

Exchange = Enum(
    'Exchange', [
        'AUS',
        'UK',
    ]
)

Endpoint = Enum(
    'Endpoint', [
        'Betting',
        'Account',
    ]
)

Wallet = Enum(
    'Wallet', [
        'UK',
        'AUSTRALIAN',
    ]
)

IncludeItem = Enum(
    'IncludeItem', [
        'ALL',
        'DEPOSITS_WITHDRAWALS',
        'EXCHANGE',
        'POKER_ROOM',
    ]
)

ItemClass = Enum(
    'ItemClass', [
        'UNKNOWN',
    ]
)



# -*- coding: utf-8 -*-

from schematics import types
from schematics.exceptions import ConversionError
from schematics.exceptions import ValidationError


class DateTimeType(types.DateTimeType):

    DEFAULT_FORMATS = (
        '%Y-%m-%dT%H:%M:%S.%f',
        '%Y-%m-%dT%H:%M:%S.%fZ',
        '%Y-%m-%dT%H:%M:%S',
    )


class EnumType(types.BaseType):

    MESSAGES = {'choices': u'Value must belong to enum {0}.'}

    def __init__(self, enum, *args, **kwargs):
        super(EnumType, self).__init__(*args, **kwargs)
        self.enum = enum

    def _to_name(self, value, error):
        if isinstance(value, self.enum):
            return value.name
        try:
            return self.enum[value].name
        except KeyError:
            message = self.messages['choices'].format(self.enum.__name__)
            raise error(message)

    def validate_choice(self, value):
        self._to_name(value, ValidationError)

    def to_native(self, value, context=None):
        return self._to_name(value, ConversionError)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from schematics.types import IntType
from schematics.types import LongType
from schematics.types import FloatType
from schematics.types import StringType
from schematics.types import BooleanType
from schematics.types.compound import DictType
from schematics.types.compound import ListType
from schematics.types.compound import ModelType

from betfair.meta.types import EnumType
from betfair.meta.types import DateTimeType
from betfair.meta.models import BetfairModel

from betfair import constants


class Event(BetfairModel):
    id = StringType()
    name = StringType()
    country_code = StringType()
    timezone = StringType()
    venue = StringType()
    open_date = DateTimeType()


class MarketDescription(BetfairModel):
    persistence_enabled = BooleanType(required=True)
    bsp_market = BooleanType(required=True)
    market_time = DateTimeType(required=True)
    suspend_time = DateTimeType(required=True)
    settle_time = DateTimeType()
    betting_type = EnumType(constants.MarketBettingType, required=True)
    turn_in_play_enabled = BooleanType(required=True)
    market_type = StringType(required=True)
    regulator = StringType(required=True)
    market_base_rate = FloatType(required=True)
    discount_allowed = BooleanType(required=True)
    wallet = StringType()
    rules = StringType()
    rules_has_date = BooleanType()
    clarifications = StringType()
    each_way_divisor = FloatType()


class RunnerCatalog(BetfairModel):
    selection_id = IntType(required=True)
    runner_name = StringType(required=True)
    handicap = FloatType(required=True)
    sort_priority = IntType(required=True)
    metadata = DictType(StringType)


class EventType(BetfairModel):
    id = StringType()
    name = StringType()


class Competition(BetfairModel):
    id = StringType()
    name = StringType()


class MarketCatalogue(BetfairModel):
    market_id = StringType()
    market_name = StringType()
    market_start_time = DateTimeType()
    description = ModelType(MarketDescription)
    total_matched = FloatType()
    runners = ListType(ModelType(RunnerCatalog))
    event_type = ModelType(EventType)
    competition = ModelType(Competition)
    event = ModelType(Event)


class TimeRange(BetfairModel):
    from_ = DateTimeType(deserialize_from='from', serialized_name='from')
    to = DateTimeType()


class MarketFilter(BetfairModel):
    text_query = StringType()
    exchange_ids = StringType()
    event_type_ids = ListType(StringType)
    event_ids = ListType(StringType)
    competition_ids = ListType(StringType)
    market_ids = ListType(StringType)
    venues = ListType(StringType)
    bsp_only = BooleanType()
    turn_in_play_enabled = BooleanType()
    in_play_only = BooleanType()
    market_betting_types = ListType(EnumType(constants.MarketBettingType))
    market_countries = ListType(StringType)
    market_type_codes = ListType(StringType)
    market_start_time = ModelType(TimeRange)
    with_orders = ListType(EnumType(constants.OrderStatus))


class PriceSize(BetfairModel):
    price = FloatType(required=True)
    size = FloatType(required=True)


class StartingPrices(BetfairModel):
    near_price = FloatType()
    far_price = FloatType()
    back_stake_taken = ListType(ModelType(PriceSize))
    lay_liability_taken = ListType(ModelType(PriceSize))
    actual_SP = FloatType()


class ExchangePrices(BetfairModel):
    available_to_back = ListType(ModelType(PriceSize))
    available_to_lay = ListType(ModelType(PriceSize))
    traded_volume = ListType(ModelType(PriceSize))


class Order(BetfairModel):
    bet_id = StringType(required=True)
    order_type = EnumType(constants.OrderType, required=True)
    status = EnumType(constants.OrderStatus, required=True)
    persistence_type = EnumType(constants.PersistenceType, required=True)
    side = EnumType(constants.Side, required=True)
    price = FloatType(required=True)
    size = FloatType(required=True)
    bsp_liability = BooleanType(required=True)
    placed_date = DateTimeType(required=True)
    avg_price_matched = FloatType()
    size_matched = FloatType()
    size_remaining = FloatType()
    size_lapsed = FloatType()
    size_cancelled = FloatType()
    size_voided = FloatType()


class Match(BetfairModel):
    bet_id = StringType()
    match_id = StringType()
    side = EnumType(constants.Side, required=True)
    price = FloatType(required=True)
    size = FloatType(required=True)
    match_date = DateTimeType()


class Runner(BetfairModel):
    selection_id = IntType(required=True)
    handicap = FloatType(required=True)
    status = EnumType(constants.RunnerStatus, required=True)
    adjustment_factor = FloatType()
    last_price_traded = FloatType()
    total_matched = FloatType()
    removal_date = DateTimeType()
    sp = ModelType(StartingPrices)
    ex = ModelType(ExchangePrices)
    orders = ListType(ModelType(Order))
    matches = ListType(ModelType(Match))


class MarketBook(BetfairModel):
    market_id = StringType(required=True)
    is_market_data_delayed = BooleanType(required=True)
    status = EnumType(constants.MarketStatus)
    bet_delay = IntType()
    bsp_reconciled = BooleanType()
    complete = BooleanType()
    inplay = BooleanType()
    number_of_winners = IntType()
    number_of_runners = IntType()
    number_of_active_runners = IntType()
    last_match_time = DateTimeType()
    total_matched = FloatType()
    total_available = FloatType()
    cross_matching = BooleanType()
    runners_voidable = BooleanType()
    version = FloatType()
    runners = ListType(ModelType(Runner))


class RunnerProfitAndLoss(BetfairModel):
    selection_id = IntType()
    if_win = FloatType()
    if_lose = FloatType()


class MarketProfitAndLoss(BetfairModel):
    market_id = StringType()
    commission_applied = FloatType()
    profit_and_losses = ListType(ModelType(RunnerProfitAndLoss))


class ExBestOffersOverrides(BetfairModel):
    best_prices_depth = IntType()
    rollup_model = EnumType(constants.RollupModel)
    rollup_limit = IntType()
    rollup_liability_threshold = FloatType()
    rollup_liability_factor = IntType()


class PriceProjection(BetfairModel):
    price_data = ListType(EnumType(constants.PriceData))
    ex_best_offers_overrides = ModelType(ExBestOffersOverrides)
    virtualise = BooleanType()
    rollover_stakes = BooleanType()


class LimitOrder(BetfairModel):
    size = FloatType(required=True)
    price = FloatType(required=True)
    persistence_type = EnumType(constants.PersistenceType, required=True)


class LimitOnCloseOrder(BetfairModel):
    liability = FloatType(required=True)
    price = FloatType(required=True)


class MarketOnCloseOrder(BetfairModel):
    liability = FloatType(required=True)


# Results

class CompetitionResult(BetfairModel):

    competition = ModelType(Competition)
    market_count = IntType()
    competition_region = StringType()


class CountryCodeResult(BetfairModel):
    country_code = StringType()
    market_count = IntType()


class EventResult(BetfairModel):

    event = ModelType(Event)
    market_count = IntType()


class EventTypeResult(BetfairModel):

    event_type = ModelType(EventType)
    market_count = IntType()


class MarketTypeResult(BetfairModel):
    market_type = StringType()
    market_count = IntType()


class TimeRangeResult(BetfairModel):

    time_range = ModelType(TimeRange)
    market_count = IntType()


class VenueResult(BetfairModel):
    venue = StringType()
    market_count = IntType()


# Instructions

class PlaceInstruction(BetfairModel):
    order_type = EnumType(constants.OrderType, required=True)
    selection_id = IntType(required=True)
    handicap = FloatType()
    side = EnumType(constants.Side, required=True)
    limit_order = ModelType(LimitOrder)
    limit_on_close_order = ModelType(LimitOnCloseOrder)
    market_on_close_order = ModelType(MarketOnCloseOrder)


class CancelInstruction(BetfairModel):
    bet_id = StringType(required=True)
    size_reduction = FloatType()


class ReplaceInstruction(BetfairModel):
    bet_id = StringType(required=True)
    new_price = FloatType(required=True)


class UpdateInstruction(BetfairModel):
    bet_id = StringType(required=True)
    new_persistence_type = EnumType(constants.PersistenceType, required=True)


# Summary reports

class CurrentOrderSummary(BetfairModel):
    bet_id = StringType(required=True)
    market_id = StringType(required=True)
    selection_id = IntType(required=True)
    handicap = FloatType(required=True)
    price_size = ModelType(PriceSize, required=True)
    bsp_liability = FloatType(required=True)
    side = EnumType(constants.Side, required=True)
    status = EnumType(constants.OrderStatus, required=True)
    persistence_type = EnumType(constants.PersistenceType, required=True)
    order_type = EnumType(constants.OrderType, required=True)
    placed_date = DateTimeType(required=True)
    matched_date = DateTimeType()
    average_price_matched = FloatType()
    size_matched = FloatType()
    size_remaining = FloatType()
    size_lapsed = FloatType()
    size_cancelled = FloatType()
    size_voided = FloatType()
    regulator_auth_code = StringType()
    regulator_code = StringType()


class CurrentOrderSummaryReport(BetfairModel):
    current_orders = ListType(ModelType(CurrentOrderSummary), required=True)
    more_available = BooleanType(required=True)


class ItemDescription(BetfairModel):
    event_type_desc = StringType()
    event_desc = StringType()
    market_desc = StringType()
    market_start_Time = DateTimeType()
    runner_desc = StringType()
    number_of_winners = IntType()


class ClearedOrderSummary(BetfairModel):
    event_type_id = StringType()
    event_id = StringType()
    market_id = StringType()
    selection_id = IntType()
    handicap = FloatType()
    bet_id = StringType()
    placed_date = DateTimeType()
    persistence_type = EnumType(constants.PersistenceType)
    order_type = EnumType(constants.OrderType)
    side = EnumType(constants.Side)
    item_description = ModelType(ItemDescription)
    price_requested = FloatType()
    settled_date = DateTimeType()
    bet_count = IntType()
    commission = FloatType()
    price_matched = FloatType()
    price_reduced = BooleanType()
    size_settled = FloatType()
    profit = FloatType()
    size_cancelled = FloatType()


class ClearedOrderSummaryReport(BetfairModel):
    cleared_orders = ListType(ModelType(ClearedOrderSummary), required=True)
    more_available = BooleanType(required=True)


# Instruction reports

class BaseInstructionReport(BetfairModel):
    status = EnumType(constants.InstructionReportStatus, required=True)
    error_code = EnumType(constants.InstructionReportErrorCode)


class PlaceInstructionReport(BaseInstructionReport):
    instruction = ModelType(PlaceInstruction, required=True)
    bet_id = StringType()
    placed_date = DateTimeType()
    average_price_matched = FloatType()
    size_matched = FloatType()


class CancelInstructionReport(BaseInstructionReport):
    instruction = ModelType(CancelInstruction)
    size_cancelled = FloatType(required=True)
    cancelled_date = DateTimeType()


class ReplaceInstructionReport(BaseInstructionReport):
    cancel_instruction_report = ModelType(CancelInstructionReport)
    place_instruction_report = ModelType(PlaceInstructionReport)


class UpdateInstructionReport(BaseInstructionReport):
    instruction = ModelType(UpdateInstruction, required=True)


# Execution reports

class BaseExecutionReport(BetfairModel):
    customer_ref = StringType()
    status = EnumType(constants.ExecutionReportStatus, required=True)
    error_code = EnumType(constants.ExecutionReportErrorCode)
    market_id = StringType()


class PlaceExecutionReport(BaseExecutionReport):
    instruction_reports = ListType(ModelType(PlaceInstructionReport))


class CancelExecutionReport(BaseExecutionReport):
    instruction_reports = ListType(ModelType(CancelInstructionReport))


class ReplaceExecutionReport(BaseExecutionReport):
    instruction_reports = ListType(ModelType(ReplaceInstructionReport))


class UpdateExecutionReport(BaseExecutionReport):
    instruction_reports = ListType(ModelType(UpdateInstructionReport))


# Accounts

class AccountFundsResponse(BetfairModel):
    available_to_bet_balance = FloatType()
    exposure = FloatType()
    retained_commission = FloatType()
    exposure_limit = FloatType()
    discount_rate = FloatType()
    points_balance = IntType()
    wallet = EnumType(constants.Wallet)


class StatementLegacyData(BetfairModel):
    avg_price = FloatType()
    bet_size = FloatType()
    bet_type = StringType()
    bet_category_type = StringType()
    commission_rate = StringType()
    event_id = LongType()
    event_type_id = LongType()
    full_market_name = StringType()
    gross_bet_amount = FloatType()
    market_name = StringType()
    market_type = StringType()
    placed_date = DateTimeType()
    selection_id = LongType()
    selection_name = StringType()
    start_date = DateTimeType()
    transaction_type = StringType()
    transaction_id = LongType()
    win_lose = StringType()


class StatementItem(BetfairModel):
    ref_id = StringType()
    item_date = DateTimeType()
    amount = FloatType()
    balance = FloatType()
    item_class = EnumType(constants.ItemClass)
    item_class_data = DictType(StringType)
    legacy_data = ModelType(StatementLegacyData)


class AccountDetailsResponse(BetfairModel):
    currency_code = StringType()
    first_name = StringType()
    last_name = StringType()
    locale_code = StringType()
    region = StringType()
    timezone = StringType()
    discount_rate = FloatType()
    points_balance = IntType()
    country_code = StringType()


class AccountStatementReport(BetfairModel):
    account_statement = ListType(ModelType(StatementItem))
    more_available = BooleanType()


class CurrencyRate(BetfairModel):
    currency_code = StringType()
    rate = FloatType()


class TransferResponse(BetfairModel):
    transaction_id = StringType()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import os
import json
import itertools
import collections

import requests
from six.moves import http_client as httplib
from six.moves import urllib_parse as urllib

from betfair import utils
from betfair import models
from betfair import exceptions


IDENTITY_URLS = collections.defaultdict(
    lambda: 'https://identitysso.betfair.com/api/',
    italy='https://identitysso.betfair.it/api/',
)

API_URLS = collections.defaultdict(
    lambda: 'https://api.betfair.com/exchange/betting/json-rpc/v1',
    australia='https://api-au.betfair.com/exchange/betting/json-rpc/v1',
)


class Betfair(object):
    """Betfair API client.

    :param str app_key: Optional application identifier
    :param str cert_file: Path to self-signed SSL certificate file(s); may be
        a *.pem file or a tuple of (*.crt, *.key) files
    :param str content_type: Optional content type
    :param str locale: Optional location ("australia", "italy", etc.)
    :param Session session: Optional Requests session
    :param int timeout: Optional timeout duration (seconds)
    """
    def __init__(self, app_key, cert_file, content_type='application/json', locale=None,
                 session=None, timeout=None):
        self.app_key = app_key
        self.cert_file = cert_file
        self.content_type = content_type
        self.locale = locale
        self.session = session or requests.Session()
        self.session_token = None
        self.timeout = timeout

    @property
    def identity_url(self):
        return IDENTITY_URLS[self.locale]

    @property
    def api_url(self):
        return API_URLS[self.locale]

    @property
    def headers(self):
        return {
            'X-Application': self.app_key,
            'X-Authentication': self.session_token,
            'Content-Type': self.content_type,
            'Accept': 'application/json',
        }

    def make_auth_request(self, method):
        response = self.session.post(
            os.path.join(self.identity_url, method),
            headers=self.headers,
            timeout=self.timeout,
        )
        utils.check_status_code(response)
        data = response.json()
        if data.get('status') != 'SUCCESS':
            raise exceptions.AuthError(response, data)

    def make_api_request(self, base, method, params, codes=None, model=None):
        payload = utils.make_payload(base, method, params)
        response = self.session.post(
            self.api_url,
            data=json.dumps(payload, cls=utils.BetfairEncoder),
            headers=self.headers,
            timeout=self.timeout,
        )
        utils.check_status_code(response, codes=codes)
        result = utils.result_or_error(response)
        return utils.process_result(result, model)

    # Authentication methods

    def login(self, username, password):
        """Log in to Betfair. Sets `session_token` if successful.

        :param str username: Username
        :param str password: Password
        :raises: BetfairLoginError
        """
        response = self.session.post(
            os.path.join(self.identity_url, 'certlogin'),
            cert=self.cert_file,
            data=urllib.urlencode({
                'username': username,
                'password': password,
            }),
            headers={
                'X-Application': self.app_key,
                'Content-Type': 'application/x-www-form-urlencoded',
            },
            timeout=self.timeout,
        )
        utils.check_status_code(response, [httplib.OK])
        data = response.json()
        if data.get('loginStatus') != 'SUCCESS':
            raise exceptions.LoginError(response, data)
        self.session_token = data['sessionToken']

    @utils.requires_login
    def keep_alive(self):
        """Reset session timeout.

        :raises: AuthError
        """
        self.make_auth_request('keepAlive')

    @utils.requires_login
    def logout(self):
        """Log out and clear `session_token`.

        :raises: AuthError
        """
        self.make_auth_request('logout')
        self.session_token = None

    # Bet query methods

    @utils.requires_login
    def list_event_types(self, filter=None, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:
        """
        filter = filter or models.MarketFilter()
        return self.make_api_request(
            'Sports',
            'listEventTypes',
            utils.get_kwargs(locals()),
            model=models.EventTypeResult,
        )

    @utils.requires_login
    def list_competitions(self, filter=None, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:
        """
        filter = filter or models.MarketFilter()
        return self.make_api_request(
            'Sports',
            'listCompetitions',
            utils.get_kwargs(locals()),
            model=models.CompetitionResult,
        )

    @utils.requires_login
    def list_time_ranges(self, granularity, filter=None):
        """

        :param TimeGranularity granularity:
        :param MarketFilter filter:
        """
        filter = filter or models.MarketFilter()
        return self.make_api_request(
            'Sports',
            'listTimeRanges',
            utils.get_kwargs(locals()),
            model=models.TimeRangeResult,
        )

    @utils.requires_login
    def list_events(self, filter=None, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:
        """
        filter = filter or models.MarketFilter()
        return self.make_api_request(
            'Sports',
            'listEvents',
            utils.get_kwargs(locals()),
            model=models.EventResult,
        )

    @utils.requires_login
    def list_market_types(self, filter=None, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:
        """
        filter = filter or models.MarketFilter()
        return self.make_api_request(
            'Sports',
            'listMarketTypes',
            utils.get_kwargs(locals()),
            model=models.MarketTypeResult,
        )

    @utils.requires_login
    def list_countries(self, filter=None, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:
        """
        filter = filter or models.MarketFilter()
        return self.make_api_request(
            'Sports',
            'listCountries',
            utils.get_kwargs(locals()),
            model=models.CountryCodeResult,
        )

    @utils.requires_login
    def list_venues(self, filter=None, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:
        """
        filter = filter or models.MarketFilter()
        return self.make_api_request(
            'Sports',
            'listCountries',
            utils.get_kwargs(locals()),
            model=models.VenueResult,
        )

    @utils.requires_login
    def list_market_catalogue(
            self, filter=None, max_results=100, market_projection=None, locale=None,
            sort=None):
        """

        :param MarketFilter filter:
        :param int max_results:
        :param list market_projection:
        :param MarketSort sort:
        :param str locale:
        """
        filter = filter or models.MarketFilter()
        return self.make_api_request(
            'Sports',
            'listMarketCatalogue',
            utils.get_kwargs(locals()),
            model=models.MarketCatalogue,
        )

    @utils.requires_login
    def list_market_book(
            self, market_ids, price_projection=None, order_projection=None,
            match_projection=None, currency_code=None, locale=None):
        """

        :param list market_ids: List of market IDs
        :param PriceProjection price_projection:
        :param OrderProjection order_projection:
        :param MatchProjection match_projection:
        :param str currency_code:
        :param str locale:
        """
        return self.make_api_request(
            'Sports',
            'listMarketBook',
            utils.get_kwargs(locals()),
            model=models.MarketBook,
        )

    @utils.requires_login
    def list_market_profit_and_loss(
            self, market_ids, include_settled_bets=False,
            include_bsp_bets=None, net_of_commission=None):
        """Retrieve profit and loss for a given list of markets.

        :param list market_ids: List of markets to calculate profit and loss
        :param bool include_settled_bets: Option to include settled bets
        :param bool include_bsp_bets: Option to include BSP bets
        :param bool net_of_commission: Option to return profit and loss net of
            users current commission rate for this market including any special
            tariffs
        """
        return self.make_api_request(
            'Sports',
            'listMarketProfitAndLoss',
            utils.get_kwargs(locals()),
            model=models.MarketProfitAndLoss,
        )

    # Chunked iterators for list methods

    def iter_list_market_book(self, market_ids, chunk_size, **kwargs):
        """Split call to `list_market_book` into separate requests.

        :param list market_ids: List of market IDs
        :param int chunk_size: Number of records per chunk
        :param dict kwargs: Arguments passed to `list_market_book`
        """
        return itertools.chain(*(
            self.list_market_book(market_chunk, **kwargs)
            for market_chunk in utils.get_chunks(market_ids, chunk_size)
        ))

    def iter_list_market_profit_and_loss(
            self, market_ids, chunk_size, **kwargs):
        """Split call to `list_market_profit_and_loss` into separate requests.

        :param list market_ids: List of market IDs
        :param int chunk_size: Number of records per chunk
        :param dict kwargs: Arguments passed to `list_market_profit_and_loss`
        """
        return itertools.chain(*(
            self.list_market_profit_and_loss(market_chunk, **kwargs)
            for market_chunk in utils.get_chunks(market_ids, chunk_size)
        ))

    # Betting methods

    @utils.requires_login
    def list_current_orders(
            self, bet_ids=None, market_ids=None, order_projection=None,
            date_range=None, order_by=None, sort_dir=None, from_record=None,
            record_count=None):
        """

        :param bet_ids:
        :param market_ids:
        :param order_projection:
        :param date_range:
        :param order_by:
        :param sort_dir:
        :param from_record:
        :param record_count:
        """
        return self.make_api_request(
            'Sports',
            'listCurrentOrders',
            utils.get_kwargs(locals()),
            model=models.CurrentOrderSummaryReport,
        )

    @utils.requires_login
    def list_cleared_orders(
            self, bet_status, event_type_ids, event_ids, market_ids,
            runner_ids, bet_ids, side, settled_date_range, group_by,
            include_item_description, locale, from_record, record_count):
        """

        :param bet_status:
        :param event_type_ids:
        :param event_ids:
        :param market_ids:
        :param runner_ids:
        :param bet_ids:
        :param side:
        :param settled_date_range:
        :param group_by:
        :param include_item_description:
        :param locale:
        :param from_record:
        :param record_count:
        """
        return self.make_api_request(
            'Sports',
            'listClearedOrders',
            utils.get_kwargs(locals()),
            model=models.ClearedOrderSummaryReport,
        )

    @utils.requires_login
    def place_orders(self, market_id, instructions, customer_ref=None):
        """Place new orders into market. This operation is atomic in that all
        orders will be placed or none will be placed.

        :param str market_id: The market id these orders are to be placed on
        :param list instructions: List of `PlaceInstruction` objects
        :param str customer_ref: Optional order identifier string
        """
        return self.make_api_request(
            'Sports',
            'placeOrders',
            utils.get_kwargs(locals()),
            model=models.PlaceExecutionReport,
        )

    @utils.requires_login
    def cancel_orders(self, market_id, instructions, customer_ref=None):
        """Cancel all bets OR cancel all bets on a market OR fully or
        partially cancel particular orders on a market.

        :param str market_id: If not supplied all bets are cancelled
        :param list instructions: List of `CancelInstruction` objects
        :param str customer_ref: Optional order identifier string
        """
        return self.make_api_request(
            'Sports',
            'cancelOrders',
            utils.get_kwargs(locals()),
            model=models.CancelInstructionReport,
        )

    @utils.requires_login
    def replace_orders(self, market_id, instructions, customer_ref=None):
        """This operation is logically a bulk cancel followed by a bulk place.
        The cancel is completed first then the new orders are placed.

        :param str market_id: The market id these orders are to be placed on
        :param list instructions: List of `ReplaceInstruction` objects
        :param str customer_ref: Optional order identifier string
        """
        return self.make_api_request(
            'Sports',
            'replaceOrders',
            utils.get_kwargs(locals()),
            model=models.ReplaceExecutionReport,
        )

    @utils.requires_login
    def update_orders(self, market_id, instructions, customer_ref=None):
        """Update non-exposure changing fields.

        :param str market_id: The market id these orders are to be placed on
        :param list instructions: List of `UpdateInstruction` objects
        :param str customer_ref: Optional order identifier string
        """
        return self.make_api_request(
            'Sports',
            'updateOrders',
            utils.get_kwargs(locals()),
            model=models.UpdateExecutionReport,
        )

    @utils.requires_login
    def get_account_funds(self, wallet=None):
        """Get available to bet amount.

        :param Wallet wallet: Name of the wallet in question
        """
        result = self.make_api_request(
            'Account',
            'getAccountFunds',
            utils.get_kwargs(locals()),
            model=models.AccountFundsResponse,
        )

    @utils.requires_login
    def get_account_statement(
            self, locale=None, from_record=None, record_count=None,
            item_date_range=None, include_item=None, wallet=None):
        """Get account statement.

        :param str locale: The language to be used where applicable
        :param int from_record: Specifies the first record that will be returned
        :param int record_count: Specifies the maximum number of records to be returned
        :param TimeRange item_date_range: Return items with an itemDate within this date range
        :param IncludeItem include_item: Which items to include
        :param Wallet wallte: Which wallet to return statementItems for
        """
        result = self.make_api_request(
            'Account',
            'getAccountStatement',
            utils.get_kwargs(locals()),
            model=models.AccountStatementReport,
        )

    @utils.requires_login
    def get_account_details(self):
        """Returns the details relating your account, including your discount
        rate and Betfair point balance.
        """
        result = self.make_api_request(
            'Account',
            'getAccountDetails',
            utils.get_kwargs(locals()),
            model=models.AccountDetailsResponse,
        )

    @utils.requires_login
    def list_currency_rates(self, from_currency=None):
        """Returns a list of currency rates based on given currency

        :param str from_currency: The currency from which the rates are computed
        """
        result = self.make_api_request(
            'Account',
            'listCurrencyRates',
            utils.get_kwargs(locals()),
            model=models.CurrencyRate,
        )

    @utils.requires_login
    def transfer_funds(self, from_, to, amount):
        """Transfer funds between the UK Exchange and Australian Exchange wallets.

        :param Wallet from_: Source wallet
        :param Wallet to: Destination wallet
        :param float amount: Amount to transfer
        """
        result = self.make_api_request(
            'Account',
            'transferFunds',
            utils.get_kwargs(locals()),
            model=models.TransferResponse,
        )



# -*- coding: utf-8 -*-

import functools

import six
import inflection


def convert_value(value, converter=None):
    return converter(value) if converter else value


def convert_dict(data, key_converter=None, value_converter=None):
    return {
        convert_value(key, key_converter): convert_value(value, value_converter)
        for key, value in six.iteritems(data)
    }


camelize = functools.partial(inflection.camelize, uppercase_first_letter=False)
serialize_dict = functools.partial(convert_dict, key_converter=camelize)



# -*- coding: utf-8 -*-

import os
import re
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


INSTALL_REQUIRES = [
    'six>=1.9.0',
    'enum34>=1.0.4',
    'invoke>=0.10.1',
    'requests>=2.6.2',
    'decorator>=3.4.2',
    'inflection>=0.3.0',
    'schematics>=1.0.4',
    'python-dateutil>=2.4.2',
]
TEST_REQUIRES = [
    'pytest',
    'responses',
]


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()


class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['--verbose']
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


def find_version(fname):
    """Attempts to find the version number in the file names fname.
    Raises RuntimeError if not found.
    """
    version = ''
    with open(fname, 'r') as fp:
        reg = re.compile(r'__version__ = [\'"]([^\'"]*)[\'"]')
        for line in fp:
            m = reg.match(line)
            if m:
                version = m.group(1)
                break
    if not version:
        raise RuntimeError('Cannot find version information')
    return version


def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content


setup(
    name='betfair.py',
    version=find_version('betfair/__init__.py'),
    description='Python client for the Betfair API '
                '(https://api.developer.betfair.com/)',
    long_description=open('README.rst').read(),
    author='Joshua Carp',
    author_email='jm.carp@gmail.com',
    url='https://github.com/jmcarp/betfair.py',
    packages=find_packages(exclude=('test*', )),
    package_dir={'betfair': 'betfair'},
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    test_suite='tests',
    tests_require=TEST_REQUIRES,
    cmdclass={'test': PyTest}
)
